<?php
/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2 as     *
 *   published by the Free Software Foundation.                            *
 *                                                                         *
 *  Caching Kolab driver developed by KONSEC GmbH, sponsored by WU Wien    *
 *  copyright            : (c) 2006-2007 by Thomas Jarosch                 *
 *  email                : opensource@intra2net.com                        *
 ***************************************************************************/

require_once 'Horde/Kolab/IMAP.php';

require_once 'Horde/MIME.php';
require_once 'Horde/MIME/Part.php';
require_once 'Horde/MIME/Message.php';
require_once 'Horde/MIME/Headers.php';
require_once 'Horde/MIME/Structure.php';

// Disk based kolab IMAP cache
class Kolab_Store
{
    // high level
    var $object_type;       // Kolab object type: contact or event
    var $owner;             // folder owner
    var $folder;            // folder name

    // low-level vars
    var $imap;              // Kolab_IMAP object

    var $store_prefix;      // prefix to prepend to all file operations
    var $diskcache_cyrus_optimize;

    var $version;           // internal version of format
    var $cache_modified;    // indicates if we have to writeout the cache

    var $cache;             // cache data
    /*
        Internal cache structure:
        | version
        |
        | uidvalidity
        |
        | uidnext
        |
        | objects: key is uid (GUID)
        | ----------- hashed object data
                      |----------- uid: object id (GUID)
        |             |----------- all fields from kolab specification
        |
        | uids   Mapping between imap uid and object uids: imap uid -> object uid
                 Special: A value of "false" means we've seen the uid
                          but we deciced to ignore it in the future
    */

    var $smart_synchronized;    // flag to synchronize only once per page load

    // vars used by MIME manipulation
    var $raw_headers;
    var $raw_message;
    var $headers;
    var $mimemsg;
    var $mimeid;

    function Kolab_Store($object_type, $server, $login, $password,
                         $folder, $port=143, $tls=false)
    {
        // Check that XML converter for this object exists
        if (!class_exists("Kolab_XML_".$object_type))
            return false;

        $this->object_type = $object_type;
        $this->owner = $login;
        $this->folder = $folder;

        // TODO: These should be configurable
        $this->store_prefix = "/tmp/imapcache/";
        $this->diskcache_cyrus_optimize = true;

        $this->version = 1;
        $this->reset_cache();
        $this->smart_synchronized = false;

        $this->imap = &new Kolab_IMAP();
        return $this->imap->connect($server, $login, $password, $port, $tls);
    }

    //
    // Memory cache handling
    //
    function reset_cache()
    {
        $this->cache = array();

        $this->cache["version"] = $this->version;
        $this->cache["uidvalidity"] = -1;
        $this->cache["uidnext"] = -1;
        $this->cache["objects"] = array();
        $this->cache["uids"] = array();

        $this->cache_modified = true;
    }

    function object_uid_exists($uid)
    {
        return array_key_exists($uid, $this->cache["objects"]);
    }

    function imap_uid_exists($uid)
    {
        return array_key_exists($uid, $this->cache["uids"]);
    }

    function get_imap_uid($object_uid)
    {
        $imap_uid = array_search($object_uid, $this->cache["uids"]);
        if ($imap_uid === false)
            return false;

        return $imap_uid;
    }

    function get_object_ids()
    {
        return array_keys($this->cache["objects"]);
    }

    function get_object($object_id)
    {
        return $this->cache["objects"][$object_id];
    }
    function get_objects()
    {
        return array_values($this->cache["objects"]);
    }

    function store_in_cache(&$imap_uid, &$object_uid, &$object)
    {
        $this->cache["uids"][$imap_uid] = $object_uid;
        $this->cache["objects"][$object_uid] = $object;
        $this->cache_modified = true;
    }

    function ignore_uid(&$imap_uid)
    {
        $this->cache["uids"][$imap_uid] = false;
        $this->cache_modified = true;
    }

    function delete($object_uid)
    {
        if (!$this->object_uid_exists($object_uid))
            return false;

        // find imap uid
        $imap_uid = $this->get_imap_uid($object_uid);
        if ($imap_uid === false)
            return false;

        if ($this->imap->delete_message($imap_uid)) {
            $this->imap->expunge();

            unset($this->cache["objects"][$object_uid]);
            unset($this->cache["uids"][$imap_uid]);

            $this->cache_modified = true;

            // No need to update cache on disk
            // as the next request will refresh it anyway
            // -> speeds up multiple deletes

            return true;
        }

        return false;
    }

    function delete_all()
    {
        $uids = array_keys();
        foreach ($this->cache["uids"] as $imap_uid => $object_uid) {
            if ($object_uid === false)
                continue;

            $this->imap->delete_message($imap_uid);
            unset($this->cache["objects"][$object_uid]);
            unset($this->cache["uids"][$imap_uid]);

            $this->cache_modified = true;
        }

        imap_expunge($this->imap);
    }

    function generateUID()
    {
        do {
            $key = md5(uniqid(mt_rand(), true));
        } while($this->object_uid_exists($key));

        return $key;
    }

    //
    // Disk serialization
    //
    function compute_filename()
    {
        if ($this->diskcache_cyrus_optimize) {
            $search_prefix = "INBOX/";

            $pos = strpos($this->folder, $search_prefix);
            if ($pos !== false && $pos == 0)
                $storename = "user/".$this->owner."/".substr($this->folder, strlen($search_prefix));
            else 
                $storename = $this->folder;
        } else
            $storename = $this->owner."/".$this->folder;

        $storename = str_replace(".", "^", $storename);
        $storename = str_replace("\0", "", $storename);

        $full_path = $this->store_prefix.$storename.".imapcache";
        return $full_path;
    }

    function cache_load()
    {
        $filename = $this->compute_filename();

        if (!is_readable($filename))
            return false;

        $this->cache = unserialize(file_get_contents($filename));

        // Delete disc cache if it's from an old version
        if ($this->cache["version"] != $this->version) {
            $this->reset_cache();
        } else
            $this->cache_modified = false;

        return true;
    }

    function cache_store($force=false)
    {
        if ($this->cache_modified || $force) {
            $filename = $this->compute_filename();

            if (!$this->mkdirhier(dirname($filename))) {
                return;
            }

            $tmpname = tempnam(dirname($this->store_prefix), 'imapcache');

            $fp = fopen($tmpname, 'w');
            if(!$fp)
                return false;
            if (fwrite($fp, serialize($this->cache)) === false) {
                fclose ($fp);
                return;
            }

            if(!rename($tmpname, $filename)) {
                return false;
            }
            fclose($fp);

            $this->cache_modified = false;
        }
    }

    function mkdirhier($dirname)
    {
        $base = substr($dirname, 0, strrpos($dirname, '/'));

        $base = str_replace(".", "^", $base);
        if( !empty($base) && !is_dir($base) ) {
            if(!$this->mkdirhier($base))
                return false;
        }

        if(!file_exists($dirname))
            return mkdir( $dirname, 0755 );

        return true;
    }

    //
    // IMAP synchronization
    //
    function smart_synchronize()
    {
        if ($this->smart_synchronized)
            return;

        $this->cache_load();
        $this->synchronize();

        $this->smart_synchronized = true;
    }

    function synchronize()
    {
        // SELECT folder so that uidnext is up to date
        $this->imap->select($this->folder);

        $status = $this->imap->folder_status();
        $new_uids = $this->imap->get_uids();
        $old_uids = array_keys($this->cache["uids"]);

        if ($this->folder_changed($status["uidvalidity"], $status["uidnext"], $old_uids, $new_uids)) {
            $recent_uids = array_diff($new_uids, $old_uids);

            $loader_name = "Kolab_XML_".$this->object_type;
            $loader = &new $loader_name;

            $count = 0;
            foreach($recent_uids as $imap_uid) {
                // get message
                $xmltext = $this->parse_message($imap_uid, false);
                $object = $loader->load($xmltext);

                if ($object !== false)
                    $this->store_in_cache($imap_uid, $object["uid"], $object);
                else
                    $this->ignore_uid($imap_uid);

                // write out cache once in a while so if the browser times out
                // we don't have to start from the beginning.
                if ($count > 500) {
                    $count = 0;
                    $this->cache_store();
                }
                $count++;
            }

            $this->cache_store();
        }
    }

    function folder_changed($uidvalidity, $uidnext, &$old_uids, &$new_uids)
    {
        $changed = false;

        // uidvalidity changed?
        if ($uidvalidity != $this->cache["uidvalidity"]) {
            $this->reset_cache();
            $changed = true;
        }

        // uidnext changed?
        if ($uidnext != $this->cache["uidnext"])
            $changed = true;

        $this->cache["uidvalidity"] = $uidvalidity;
        $this->cache["uidnext"] = $uidnext;

        // Speed optimization: if uidnext and uidvalidity didn't change
        // and count(old_uids) == count(new_uids), the folder didn't change.
        if ($changed || count($old_uids) != count ($new_uids)) {
            // remove deleted messages from cache
            $delete_uids = array_diff($old_uids, $new_uids);
            foreach($delete_uids as $delete_uid) {
                $object_uid = $this->cache["uids"][$delete_uid];
                if ($object_uid !== false)
                    unset($this->cache["objects"][$object_uid]);
                unset($this->cache["uids"][$delete_uid]);
            }
            $this->cache_modified = true;

            $changed = true;
        }

        return $changed;
    }

    // Get IMAP message and find Kolab XML object
    function parse_message($imap_uid, $parse_headers=true)
    {
        $error = false;

        $this->raw_headers = $this->imap->get_message_header($imap_uid);
        $body = $this->imap->get_message_body($imap_uid);

        $this->raw_message = $this->raw_headers.$body;

        $this->mimemsg = &MIME_Structure::parseTextMIMEMessage($this->raw_message);
        $parts = $this->mimemsg->contentTypeMap();

        // Read in a Kolab event object, if one exists
        $this->mime_part_id = array_search("application/x-vnd.kolab.".$this->object_type, $parts);
        if ($this->mime_part_id !== false) {
                if ($parse_headers)
                    $this->headers = &MIME_Structure::parseMIMEHeaders($this->raw_headers);

                $part = $this->mimemsg->getPart($this->mime_part_id);
                $part->transferDecodeContents();

                $xml = $part->getContents();

                return $xml;
        }

        // Nothing found -> clear MIME vars
        $this->raw_headers = false;
        $this->raw_message = false;
        $this->headers = false;
        $this->mimemsg = false;
        $this->mime_part_id = false;

        return false;
    }

    function save($object,$old_object_id=null)
    {
        $saver_name = "Kolab_XML_".$this->object_type;
        $saver = &new $saver_name;

        $new_headers = new MIME_Headers();

        // update existing kolab object
        if ($old_object_id != null) {
            // check if object really exists
            if (!$this->object_uid_exists($old_object_id))
                return false;

            // get IMAP uid
            $imap_uid = $this->get_imap_uid($old_object_id);
            if ($imap_uid === false)
                return false;

            // parse email and load XML structure
            $old_xml = $this->parse_message($imap_uid);
            if ($old_xml === false || $saver->load($old_xml) === false)
                return false;

            // copy email header
            foreach($this->headers as $header => $value)
                $new_headers->addheader($header, $value);
        } else {
            $this->prepare_new_message($new_headers);
            $this->mime_part_id = false;
        }

        $xml = $saver->save($object);

        // Update mime part
        $part = new MIME_Part("application/x-vnd.kolab.".$this->object_type, $xml, NLS::getCharset());
        $part->setTransferEncoding("quoted-printable");
        $part->setName($this->object_type.".xml");

        if ($this->mime_part_id === false)
            $this->mimemsg->addPart($part);
        else
            $this->mimemsg->alterPart($this->mime_part_id, $part);

        // Update email headers
        $new_headers->addHeader('From', Auth::getAuth());
        $new_headers->addHeader('To', Auth::getAuth());
        $new_headers->addHeader('Date', date('r'));
        $new_headers->addHeader("X-Kolab-Type", "application/x-vnd.kolab.".$this->object_type);
        $new_headers->addHeader('Subject', $object["uid"]);
        $new_headers->addHeader("User-Agent", "Intra2net AG Kolab Connector 1.0");
        $new_headers->addMIMEHeaders($this->mimemsg);

        $msg = preg_replace("/\r\n|\n|\r/s", "\r\n", $new_headers->toString() . $this->mimemsg->toString(false));

        // delete old email?
        if ($old_object_id != null)
            $this->imap->delete_message($imap_uid);

        // store new email
        if (!$this->imap->append_message($msg)) {
            if ($old_object_id != null)
                $this->imap->undelete_message($imap_uid);

            return false;
        }

        // remove deleted object
        if ($old_object_id != null)
            $this->imap->expunge();

        $this->synchronize();

        return true;
    }

    function prepare_new_message()
    {
        // New email code
        $this->mimemsg = new MIME_Message();

        $kolab_text = sprintf(_("This is a Kolab Groupware object. To view this object you will need an email client that understands the Kolab Groupware format. For a list of such email clients please visit %s"),
                                'http://www.kolab.org/kolab2-clients.html');

        $part = new MIME_Part('text/plain', String::wrap($kolab_text, 76, "\r\n", NLS::getCharset()),
                                NLS::getCharset());
        $part->setTransferEncoding('quoted-printable');
        $this->mimemsg->addPart($part);
    }
}
