<?php
/***************************************************************************
 *  This program is free software; you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License version 2 as      *
 *  published by the Free Software Foundation.                             *
 *                                                                         *
 *  Kronolith driver implementation for KONSEC groupware                   *
 *                                                                         *
 *  Caching Kolab driver developed by KONSEC GmbH, sponsored by WU Wien    *
 *  copyright            : (c) 2006-2007 by Thomas Jarosch                 *
 *  email                : opensource@intra2net.com                        *
 ***************************************************************************
 * Based upon Kronolith SQL driver
 *
 * @author  Thomas Jarosch <thomas.jarosch@intra2net.com>
 * @since   Kronolith 0.3
 * @package Kronolith
 */

require_once 'Horde/Kolab/XML_event.php';
require_once 'Horde/Kolab/Store.php';
require_once 'Horde/Kolab/Util.php';

class Kronolith_Driver_kolab extends Kronolith_Driver
{
    // internal cache of Kronolith_Event_kolab. UID is key
    var $events_cache;

    var $store;
    var $synchronized;

    function open($calendar)
    {
        $this->_calendar = $calendar;

        $this->synchronized = false;
        $this->events_cache = array();

        Horde::logMessage("open() for calendar ".$this->_calendar, __FILE__, __LINE__, PEAR_LOG_ERR);
    }

    // we delay initial synchronization to the first use
    // so having multiple address books doesn't add to the total latency.
    // This function must be called before all internal driver functions
    function synchronize($force=false)
    {
        if ($this->synchronized && !$force)
            return;

        // FIXME: Get real params
        // $params = $this->getParams();
        $params = array();

        if ($this->_calendar == "kolab2") {
            $params["server"] = "127.0.0.1";
            $params["login"] = "admin";
            $params["password"] = "admin";
            $params["folder"] = "INBOX/Kalender Technik";
        } else {
            $params["server"] = "127.0.0.1";
            $params["login"] = "admin";
            $params["password"] = "admin";
            $params["folder"] = "INBOX/Kalender";
        }
/*
        $params["server"] = "172.16.1.99";
        $params["login"] = "max.mustermann";
        $params["password"] = "max.mustermann";
        $params["folder"] = "INBOX/Kalender";
*/

        $this->store = new Kolab_Store("event",
                                        $params["server"],
                                        $params["login"],
                                        $params["password"],
                                        $params["folder"]);

        $this->store->synchronize();

        // build internal event cache
        $this->events_cache = array();
        $events = $this->store->get_objects();
        foreach($events as $event)
            $this->events_cache[$event["uid"]] = &new Kronolith_Event_kolab($this, $event);

        $this->synchronized = true;
    }

    function listAlarms($date)
    {
        $allevents = $this->listEvents($date, null, true);
        $events = array();

        foreach ($allevents as $eventId) {
            $event = &$this->getEvent($eventId);

            if ($event->getRecurType() == KRONOLITH_RECUR_NONE) {
                $start = &new Horde_Date($event->start);
                $start->min -= $event->getAlarm();
                $start->correct();
                if ($start->compareDateTime($date) <= 0 &&
                    $date->compareDateTime($event->end) <= -1) {
                    $events[] = $eventId;
                }
            } else {
                if ($next = $event->nextRecurrence($date)) {
                    $start = &new Horde_Date($next);
                    $start->min -= $event->getAlarm();
                    $start->correct();
                    $end = &new Horde_Date(array('year' => $next->year,
                                                 'month' => $next->month,
                                                 'mday' => $next->mday,
                                                 'hour' => $event->end->hour,
                                                 'min' => $event->end->min,
                                                 'sec' => $event->end->sec));
                    if ($start->compareDateTime($date) <= 0 &&
                        $date->compareDateTime($end) <= -1) {
                        $events[] = $eventId;
                    }
                }
            }
        }

        return is_array($events) ? $events : array();
    }

    /**
     * Checks if the event's UID already exists and returns all event
     * ids with that UID.
     *
     * @param string $uid          The event's uid.
     * @param string $calendar_id  Calendar to search in.
     *
     * @return string|boolean  Returns a string with event_id or false if
     *                         not found.
     */
    function exists($uid, $calendar_id = null)
    {
        $this->synchronize();

        if ($this->store->object_uid_exists($uid))
            return $uid;
        else
            return false;
    }

    /**
     * Lists all events in the time range, optionally restricting
     * results to only events with alarms.
     *
     * @param Horde_Date $startInterval  Start of range date object.
     * @param Horde_Date $endInterval    End of range data object.
     * @param boolean $hasAlarm          Only return events with alarms?
     *                                   Defaults to all events.
     *
     * @return array  Events in the given time range.
     */
    function listEvents($startDate = null, $endDate = null, $hasAlarm = false)
    {
        $this->synchronize();

        if (is_null($startDate))
            $startDate = &new Horde_Date(array('mday' => 1, 'month' => 1, 'year' => 0000));
        if (is_null($endDate))
            $endDate = &new Horde_Date(array('mday' => 31, 'month' => 12, 'year' => 9999));

        $ids = array();

        foreach($this->events_cache as $event) {
            if ($hasAlarm && !$event->getAlarm())
                continue;

            $keep_event = false;

            // check normal event date
            if ($startDate->compareDate($event->start) <= 0
                && $endDate->compareDate($event->end) >= 0)
            {
                $keep_event = true;
            }

            // do recurrence expansion if not keeping anyway
            if (!$keep_event && $event->getRecurType() != KRONOLITH_RECUR_NONE) {
                $next = $event->nextRecurrence($startDate);

                if ($next !== false && $startDate->compareDate($next) <= 0
                    && $endDate->compareDate($next) >= 0)
                {
                    $keep_event = true;
                }
            }

            if ($keep_event)
                $ids[] = $event->getUID();
        }
/*
        echo "<pre>\n";
        echo "listEvents():\n";
        if (!is_null($startDate)) {
            echo "start: ";
            $startDate->dump();
        }
        if (!is_null($endDate)) {
            echo "end: ";
            $endDate->dump();
        }
        echo "hasAlarm: $hasAlarm<br>\n";
        echo "</pre>\n";

        echo "<pre>IDs:\n";
        print_r($ids);
        echo "</pre>\n";
*/
        return $ids;
    }

    function &getEvent($eventId = null)
    {
        $this->synchronize();

        if (is_null($eventId)) {
            $event = &new Kronolith_Event_kolab($this);
            return $event;
        }

        if (array_key_exists($eventId, $this->events_cache))
            return $this->events_cache[$eventId];

        print_r($this->events_cache);

        return PEAR::raiseError(_("Event not found: $eventId in ".$this->imapserver));
    }

    function &getByUID($uid, $getAll = false)
    {
        return $this->getEvent($uid);
    }

    /**
     * Saves an event in the backend.
     * If it is a new event, it is added, otherwise the event is updated.
     *
     * @param Kronolith_Event $event  The event to save.
     */
    function saveEvent(&$event)
    {
        $this->synchronize();

        $uid = $event->getUID();
        if ($uid == null)
            $uid = $this->store->generateUID();

        $attributes = $event->toDriver($uid);

        if ($event->isStored() || $event->exists()) {
            if (!$this->store->save($attributes, $attributes["uid"]))
                return false;
        } else {
            if (!$this->store->save($attributes))
                return false;
        }

        // refresh IMAP cache
        $this->synchronize(true);

        return true;
    }

    /**
     * Move an event to a new calendar.
     *
     * @param string $eventId      The event to move.
     * @param string $newCalendar  The new calendar.
     */
    function move($eventId, $newCalendar)
    {
        $this->synchronize();

        Horde::logMessage("move() event $eventId from ".$this->_calendar." to $newCalendar", __FILE__, __LINE__, PEAR_LOG_ERR);

        // Get imap UID
        $imap_uid = $this->store->get_imap_uid($eventId);
        if ($imap_uid === false) {
            Horde::logMessage("get of IMAP uid for event $eventId failed", __FILE__, __LINE__, PEAR_LOG_ERR);
            return false;
        }

        global $kronolith_shares;
        $target = &$kronolith_shares->getShare($newCalendar);
        $folder = $target->get('kolab_folder');

        Horde::logMessage("target folder: $folder", __FILE__, __LINE__, PEAR_LOG_ERR);

        // Moving is evil as it assumes all events are stored in the same storage backend
        // This needs to be fixed in Kronolith itself
        if ($this->store->imap->move_message($imap_uid, $folder)) {
            $this->synchronize();
            return true;
        }

        return false;
    }

    /**
     * Delete a calendar and all its events.
     *
     * @param string $calendar  The name of the calendar to delete.
     *
     * @return mixed  True or a PEAR_Error on failure.
     */
    function delete($calendar)
    {
        // this is done by IMAP folder rename
        return false;
    }

    /**
     * Rename a calendar.
     *
     * @param string $from  The current name of the calendar.
     * @param string $to    The new name of the calendar.
     *
     * @return mixed  True or a PEAR_Error on failure.
     */
    function rename($from, $to)
    {
        // this is done by IMAP folder rename
        return false;
    }

    /**
     * Delete an event.
     *
     * @param string $eventId  The ID of the event to delete.
     *
     * @return mixed  True or a PEAR_Error on failure.
     */
    function deleteEvent($eventId)
    {
        $this->synchronize();

        if (!$this->store->object_uid_exists($eventId))
            return false;

        $event = &$this->getEvent($eventId);

        if ($this->store->delete($eventId)) {
            unset($this->events_cache[$eventId]);

            // Notify about the deleted event.
            $result = Kronolith::sendNotification($event, 'delete');
            if (is_a($result, 'PEAR_Error')) {
                Horde::logMessage($result, __FILE__, __LINE__, PEAR_LOG_ERR);
                return false;
            }

            return true;
        }

        return false;
    }

    function close()
    {
        Horde::logMessage("close() for calendar ".$this->_calendar, __FILE__, __LINE__, PEAR_LOG_ERR);

        return true;
    }
}

/**
 * @package Kronolith
 */
class Kronolith_Event_kolab extends Kronolith_Event {
    function fromDriver($event)
    {
        // $driver = &$this->getDriver();
        $this->eventID = $event["uid"];
        $this->setUID($event["uid"]);
        $this->title = $event["summary"];
        $this->description = $event["body"];
        $this->location = $event["location"];
        $this->category = $event["categories"];

        $this->creatorID = $event["organizer"];

        $this->alarm = $event["alarm"];

        $this->start = new Horde_Date($event["start-date"]);
        $this->end = new Horde_Date($event["end-date"]);
        $this->durMin = ($this->end->timestamp() - $this->start->timestamp()) / 60;

        switch ($event["show-time-as"]) {
        case 'free':
            $this->status = KRONOLITH_STATUS_CANCELLED;
            break;

        case 'tentative':
            $this->status = KRONOLITH_STATUS_TENTATIVE;
            break;

        case 'busy':
        case 'outofoffice':
        default:
            $this->status = KRONOLITH_STATUS_CONFIRMED;
        }

        // recurrence
        $this->recurEnd = null;
        $this->recurType = KRONOLITH_RECUR_NONE;

        // we fake recurrences by count
        // by calculating all recurrences before
        // adding exceptions and set an end date
        $fake_recurrence_end = false;
        $recurrence_count = 0;

        if (isset($event["recurrence"])) {
            $recurrence = $event["recurrence"];

            $this->recurInterval = $recurrence["interval"];

            switch ($recurrence["cycle"]) {
            case 'daily':
                $this->recurType = KRONOLITH_RECUR_DAILY;
                break;

            case 'weekly':
                $this->recurType = KRONOLITH_RECUR_WEEKLY;

                $mask = 0;
                $bits = array(
                    'monday' => HORDE_DATE_MASK_MONDAY,
                    'tuesday' => HORDE_DATE_MASK_TUESDAY,
                    'wednesday' => HORDE_DATE_MASK_WEDNESDAY,
                    'thursday' => HORDE_DATE_MASK_THURSDAY,
                    'friday' => HORDE_DATE_MASK_FRIDAY,
                    'saturday' => HORDE_DATE_MASK_SATURDAY,
                    'sunday' => HORDE_DATE_MASK_SUNDAY,
                );

                foreach ($recurrence["days"] as $day) {
                    // validity check
                    if (empty($day) || !isset($bits[$day]))
                        continue;

                    $mask |= $bits[$day];
                }

                $this->setRecurOnDay($mask);
                break;

            case 'monthly':
                switch ($recurrence["type"]) {
                case 'daynumber':
                    $this->recurType = KRONOLITH_RECUR_DAY_OF_MONTH;
                    break;

                case 'weekday':
                    $this->recurType = KRONOLITH_RECUR_WEEK_OF_MONTH;
                    break;
                }
                break;

            case 'yearly':
                $this->recurType = KRONOLITH_RECUR_YEARLY;
                break;
            }

            switch ($recurrence["range-type"]) {
            case 'number':
                $fake_recurrence_end = true;
                $recurrence_count = $recurrence["range"];
                break;

            case 'date':
                // fix of-by-one day
                $this->recurEnd = new Horde_Date($recurrence["range"] + 86400);
                break;
            }

            // calculate fake recurrence end before exclusions
            if ($fake_recurrence_end) {
                $end_date = $this->start;
                for ($count = 1; $count < $recurrence_count; $count++) {
                    $end_date = $this->nextRecurrence($end_date);
                    $end_date->mday++;
                    $end_date->correct();
                }
                $this->recurEnd = $end_date;
            }

            // recurrence exclusions
            foreach($recurrence["exclusions"] as $exclusion) {
                $date = new Horde_Date($exclusion);
                $this->addexception($date->year, $date->month, $date->mday);
            }
        }

        // attendees
        foreach($event["attendees"] as $attendee) {
            $name = $attendee["display-name"];
            $email = $attendee["smtp-address"];

            // Fake email address if none is known
            if (empty($email) && !empty($name))
                $email = $name."@no_email";
            if (empty($email))
                continue;

            $role = $attendee["role"];
            switch ($role) {
            case 'optional':
                $role = KRONOLITH_PART_OPTIONAL;
                break;

            case 'resource':
                $role = KRONOLITH_PART_NONE;
                break;

            case 'required':
            default:
                $role = KRONOLITH_PART_REQUIRED;
                break;
            }

            $status = $attendee["status"];
            switch ($status) {
            case 'accepted':
                $status = KRONOLITH_RESPONSE_ACCEPTED;
                break;

            case 'declined':
                $status = KRONOLITH_RESPONSE_DECLINED;
                break;

            case 'tentative':
                $status = KRONOLITH_RESPONSE_TENTATIVE;
                break;

            case 'none':
            default:
                $status = KRONOLITH_RESPONSE_NONE;
                break;
            }

            $this->addAttendee($email, $role, $status);
        }

        $this->initialized = true;
        $this->stored = true;
    }

    function toDriver($uid)
    {
        $event = array();
        $event["uid"] = $uid;
        $event["summary"] = $this->title;
        $event["body"]  = $this->description;
        $event["location"] = $this->location;
        $event["categories"] = $this->category;

        // FIXME!!!
        $this->creatorID = "admin";

        $event["alarm"] = $this->alarm;

        $event["start-date"] = Kolab_Util::encodeDateTime($this->start->timestamp());
        $event["end-date"] = Kolab_Util::encodeDateTime($this->end->timestamp());

        switch ($this->status) {
        case KRONOLITH_STATUS_CANCELLED:
            $event["show-time-as"] = "free";
            break;

        case KRONOLITH_STATUS_TENTATIVE:
            $event["show-time-as"] = "tentative";
            break;

        // FIXME: No mapping for outofoffice
        case KRONOLITH_STATUS_CONFIRMED:
        default:
            $event["show-time-as"] = "busy";
        }

        // recurrence
        if ($this->recurType != KRONOLITH_RECUR_NONE) {
            $recurrence = array();

            $recurrence["interval"] = $this->recurInterval;

            switch ($this->recurType) {
            case KRONOLITH_RECUR_DAILY:
                $recurrence["cycle"] = "daily";
                break;

            case KRONOLITH_RECUR_WEEKLY:
                $recurrence["cycle"] = "weekly";

                $bits = array(
                    'monday' => HORDE_DATE_MASK_MONDAY,
                    'tuesday' => HORDE_DATE_MASK_TUESDAY,
                    'wednesday' => HORDE_DATE_MASK_WEDNESDAY,
                    'thursday' => HORDE_DATE_MASK_THURSDAY,
                    'friday' => HORDE_DATE_MASK_FRIDAY,
                    'saturday' => HORDE_DATE_MASK_SATURDAY,
                    'sunday' => HORDE_DATE_MASK_SUNDAY,
                );

                $days = array();
                foreach($bits as $name => $bit) {
                    if ($this->recurOnDay($bit))
                        $days[] = $name;
                }

                $recurrence["days"] = $days;
                break;

            case KRONOLITH_RECUR_DAY_OF_MONTH:
                $recurrence["cycle"] = "monthly";
                $recurrence["type"] = "daynumber";
                $recurrence["daynumber"] = $this->start->mday;
                break;

            case KRONOLITH_RECUR_WEEK_OF_MONTH:
                $recurrence["cycle"] = "monthly";
                $recurrence["type"] = "weekday";
                $recurrence["daynumber"] = "1";

                $conv = array(
                    "sunday",
                    "monday",
                    "tuesday",
                    "wednesday",
                    "thursday",
                    "friday",
                    "saturday"
                );
                $recurrence["days"] = array ($conv[$this->start->dayOfWeek()]);
                break;

            case KRONOLITH_RECUR_YEARLY:
                $recurrence["cycle"] = "yearly";
                break;
            }

            $recurrence["range-type"] = "none";
            if ($this->recurEnd != null) {
                // fix of-by-one day
                $date = $this->recurEnd;
                $date->mday -= 1;
                $date->correct();

                $recurrence["range-type"] = "date";
                $recurrence["range"] = $date->year."-".$date->month."-".$date->mday;
            } else {
                $recurrence["range-type"] = "nothing";
                $recurrence["range"] = "";
            }

            // recurrence exclusions
            $exclusions = array();
            foreach ($this->getExceptions() as $exception) {
                if (!empty($exception)) {
                    list($year, $month, $mday) = sscanf($exception, '%04d%02d%02d');
                    $exclusions[] = "$year-$month-$mday";
                }
            }
            $recurrence["exclusions"] = $exclusions;

            $event["recurrence"] = $recurrence;
        }

        // attendee
        $event["attendees"] = array();
        $attendees = $this->getAttendees();
        foreach($attendees as $email => $attendee) {
            $new_attendee = array();
            // $new_attendee["display-name"] = "";
            $new_attendee["smtp-address"] = $email;

            switch ($attendee["attendance"]) {
            case KRONOLITH_PART_OPTIONAL:
                $new_attendee["role"] = "optional";
                break;

            case KRONOLITH_PART_NONE:
                $new_attendee["role"] = "resource";
                break;

            case KRONOLITH_PART_REQUIRED:
            default:
                $new_attendee["role"] = "required"; 
                break;
            }

            $new_attendee["request-response"] = "0";

            switch ($attendee["response"]) {
            case KRONOLITH_RESPONSE_ACCEPTED:
                $new_attendee["status"] = "accepted";
                break;

            case KRONOLITH_RESPONSE_DECLINED:
                $new_attendee["status"] = "declined";
                break;

            case KRONOLITH_RESPONSE_TENTATIVE:
                $new_attendee["status"] = "tentative";
                break;

            case KRONOLITH_RESPONSE_NONE:
            default:
                $new_attendee["status"] = "none";
                break;
            }

            $event["attendees"][] = $new_attendee;
        }

        return $event;
    }
}
