<?php
/***************************************************************************
 *  This program is free software; you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License version 2 as      *
 *  published by the Free Software Foundation.                             *
 *                                                                         *
 *  Turba directory driver implementation for KONSEC groupware             *
 *                                                                         *
 *  Caching Kolab driver developed by KONSEC GmbH, sponsored by WU Wien    *
 *  copyright            : (c) 2006-2007 by Thomas Jarosch                 *
 *  email                : opensource@intra2net.com                        *
 ***************************************************************************
 *
 * @author  Thomas Jarosch <thomas.jarosch@intra2net.com>
 * @since   Turba 2.1.2
 * @package Turba
 */

/*
Speed optimizations for large (8000+) address books:
    - Use turba keys in cache + extra capability to skip toTurbaKeys
*/

require_once 'Horde/Kolab/XML_contact.php';
require_once 'Horde/Kolab/Store.php';

class Turba_Driver_kolab extends Turba_Driver {
    var $_capabilities;
    var $store;

    function _init()
    {
        $this->_capabilities = array("delete_all" => true, "uses_turba_keys" => true);

        return true;
    }

    // we delay initial synchronization to the first use
    // so having multiple address books doesn't add to the total latency.
    // This function must be called before all internal driver functions
    function init_store()
    {
        $params = $this->getParams();

        $this->store = new Kolab_Store("contact",
                                        $params["server"],
                                        $params["login"],
                                        $params["password"],
                                        $params["folder"]);

        $this->store->smart_synchronize();
    }

    /**
     * Returns all entries - searching isn't implemented here for now. The
     * parameters are simply ignored.
     *
     * @param array $criteria  Array containing the search criteria.
     * @param array $fields    List of fields to return.
     *
     * @return array  Hash containing the search results.
     */

    function _search($criteria, $fields)
    {
        $this->init_store();

        $entries = $this->store->get_objects();

        // keep only entries matching criteria
        if (count($criteria)) {
            $new_entries = array();

            foreach($entries as $entry) {
                $keep_entry = false;

                foreach ($criteria as $key => $tests) {
                    if ($key == 'AND' || $key == 'OR') {
                        $all_matched = true;

                        foreach ($tests as $test) {
                            $field = $test["field"];
                            $needle = $test["test"];

                            if (empty($needle))
                                $keep_entry = true;
                            else if (stristr($entry[$field], $needle) !== false)
                                $keep_entry = true;
                            else
                                $all_matched = false;
                        }
                    }

                    if ($key == 'AND' && !$all_matched)
                        $keep_entry = false;
                }

                if ($keep_entry)
                    $new_entries[] = $entry;
            }

            $entries = $new_entries;
        }

        return $entries;
    }

    /**
     * Reads the given data from the preferences and returns the result's
     * fields.
     *
     * @param array $criteria  Search criteria.
     * @param string $id       Data identifier.
     * @param array $fields    List of fields to return.
     *
     * @return  Hash containing the search results.
     */
    function _read($criteria, $ids, $fields)
    {
        $this->init_store();

        $results = array();

        if (!is_array($ids))
            $ids = array($ids);

        foreach ($ids as $id)
            if ($this->store->object_uid_exists($id))
                $results[] = $this->store->get_object($id);

        return $results;
    }

    /**
     * Adds the specified object to the preferences.
     */

    function _add($attributes)
    {
        $this->init_store();

        return $this->store->save($attributes, null);
    }

    /**
     * Saves the specified object in the preferences.
     */

    function _save($object_key, $object_id, $attributes)
    {
        $this->init_store();

        if ($object_key != 'uid')
            return false;

        if (!$this->store->save($attributes, $object_id))
            return false;

        return $object_id;
    }

    /**
     * Deletes the specified object from the preferences.
     */
    function _delete($object_key, $object_id)
    {
        $this->init_store();

        if ($object_key != 'uid')
            return false;

        $this->store->delete($object_id);

        return true;
    }

    /**
     * Deletes all contacts from a specific address book.
     *
     * @param string  $sourceName  The owner_id of the address book to
     *                             delete.  If omitted, the user's default
     *                             address book will be cleared.
     *
     * @return boolean  True if the operation worked.
     */
    function _deleteAll($sourceName = '')
    {
        $this->init_store();

        $this->store->delete_all();

        return true;
    }

    /**
     * Create an object key for a new object.
     *
     * @param array $attributes  The attributes (in driver keys) of the
     *                           object being added.
     *
     * @return string  A unique ID for the new object.
     */
    function _makeKey($attributes)
    {
        $this->init_store();

        return $this->store->generateUID();
    }
}
